Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import LemmaVTMap.
Require Import LemmaInterference.
Require Import LemmaOtherWorkload.


Open Scope Z_scope.

Section Sched_type.

  Lemma busy_schedule_every_tick_is_real2:
    forall N ts te conf v,
      0 < N ->
      ts <= te ->
      last_idle_slot N N ts te conf = Some v ->
      (forall t, v + 1 <= t < te ->
            0 <= highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf (real_priqueue (Z.to_nat N) t conf) < N).
  Proof.
    intros * HN Hts Hidle t Ht.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte in *.
    clear Hte Heqnrange Hts.
    unfold last_idle_slot in Hidle.
    assert (H1: (Z.to_nat (ts + Z.of_nat nrange - ts)) = nrange).
    {
      replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
      rewrite Nat2Z.id; auto.
    }
    rewrite H1 in Hidle. clear H1.
    induction nrange.
    {
      simpl in Hidle.
      congruence.
    }
    {
      simpl in Hidle.
      
      assert (Hrealnow: highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                                 (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <> N).
      {
        destruct (zeq N
                (highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf))); auto.
        inversion Hidle.
        xomega.
      }      
      rewrite zeq_false in Hidle; auto.
      destruct (zeq t (ts + Z.of_nat nrange)) as [Heq | Hneq].
      {
        assert (0 <= highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                             (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <= N).
        {
          apply highest_not_finished_map_res_bound; try omega.
          - replace N with (Z.of_nat (Z.to_nat N)) at 1.
            apply real_priqueue_sound.
            rewrite Z2Nat.id; omega.
        }
        rewrite Heq in *.
        omega.
      }
      {
        apply IHnrange; auto.
        rewrite Nat2Z.inj_succ in Ht.
        omega.
      }
    }
  Qed.

  Lemma every_tick_is_real_last_idle_is_None:
    forall N ts te conf,
      0 < N ->
      ts <= te ->
      (forall t, ts <= t < te ->
            0 <= highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf (real_priqueue (Z.to_nat N) t conf) < N) ->      
      last_idle_slot N N ts te conf = None.
  Proof.
    intros * HN Hts Hidle.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte in *.
    clear Hte Heqnrange Hts.
    unfold last_idle_slot.
    assert (H1: (Z.to_nat (ts + Z.of_nat nrange - ts)) = nrange).
    {
      replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
      rewrite Nat2Z.id; auto.
    }
    rewrite H1. clear H1.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (Hind: forall t : Z,
              ts <= t < ts + Z.of_nat nrange ->
              0 <=
              highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                                       (real_priqueue (Z.to_nat N) t conf) < N).
      {
        intros * Ht.
        apply Hidle.
        xomega.
      }
      specialize (IHnrange Hind).
      assert (H1: ts <= ts + Z.of_nat nrange < ts + Z.of_nat (S nrange)). xomega.
      specialize (Hidle _ H1).
      rewrite zeq_false; try omega.
      apply IHnrange.
    }
  Qed.
  
  Lemma last_idle_slot_none:
    forall N ts te conf v,
      0 < N ->
      0 <= ts <= te ->
      last_idle_slot N N ts te conf = Some v ->
      last_idle_slot N N (v + 1) te conf = None.
  Proof.
    intros * HN Hts Hlast.
    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }
    apply every_tick_is_real_last_idle_is_None; auto; try omega.
    - apply busy_schedule_every_tick_is_real2 with (ts:=ts); auto.
      omega.
  Qed.

  Lemma sched_type_is_complete_when_schedule_is_busy_generalized:
    forall N conf id T C ts te todeadline,
      0 <= ts <= te ->
      1 < T ->
      0 < C ->
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      last_idle_slot N N ts te conf = None ->
      count_type_occurrence N N ts te conf id todeadline Sched_self +
      count_type_occurrence N N ts te conf id todeadline Sched_interference +
      count_type_occurrence N N ts te conf id todeadline Sched_leftover +
      count_type_occurrence N N ts te conf id todeadline Sched_overlap = te - ts.
  Proof.
    intros * Hts Htsmod HT HC Hid Hconfid Hidle.
    unfold count_type_occurrence.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (H1: te - ts = Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite H1.
    replace te with (ts + Z.of_nat nrange) in Hidle by omega.

    apply sched_type_is_complete_when_schedule_is_busy_aux with (T:=T) (C := C); auto.
    omega.
  Qed.
  
End Sched_type.



Section Sched_self.

  Lemma last_idle_is_idle:
    forall N v conf ts te,
      last_idle_slot N N ts te conf = Some v ->
      highest_not_finished_map N v (TAC_pool_at_t v (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) v conf) = N.
  Proof.
    intros * Hidle.
    unfold last_idle_slot in Hidle.
    remember (Z.to_nat (te - ts)) as nrange.
    clear Heqnrange.
    induction nrange.
    {
      simpl in Hidle.
      congruence.
    }
    {
      simpl in Hidle.
      destruct (zeq N
                (highest_not_finished_map N (ts + Z.of_nat nrange)
                   (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf))) as [Heq |Hneq].
      {
        inversion Hidle.
        rewrite H0 in *.
        auto.
      }
      {
        apply IHnrange; auto.
      }
    }
  Qed.

  Lemma ready_task_must_be_within_last_idle:
    forall N conf v ts te t i T C,
      1 < T ->
      0 <= ts <= te ->
      v + 1 <= t ->
      last_idle_slot N N ts te conf = Some v ->      
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 <= i < N ->
      vt_map_not_finished t T C (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true ->
      v + 1 <= t / T * T.
  Proof.
    intros * HT Hts Ht Hidle Hconfi Hi Hfinish.
    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }

    destruct (zle (t / T * T) v); try omega.
    exfalso.

    assert (H1: 0 <= t / T * T).
    { apply div_mul_nonneg; omega. }
    assert (H2: t / T * T <= t).
    { apply divide_mul_le; omega. }
    assert (H3: (t / T * T) mod T = 0).
    { apply Z_mod_mult. }
    assert (H4: t < t / T * T + T).
    { apply x_lt_floor_plus; omega. }
    assert (H5: v / T * T = t / T * T).
    { apply divide_mul_plus_equal2; try omega. }
    
    assert (Hfinishv: vt_map_not_finished v T C (ZMap.get i (TAC_pool_at_t v (Z.to_nat N) conf)) = true).
    {
      unfold vt_map_not_finished in *.
      rewrite H5.
        
      assert (H10: ZMap.get i (TAC_pool_at_t (t/ T*T) (Z.to_nat N) conf) (t / T * T) = ZMap.get i (TAC_pool_at_t v (Z.to_nat N) conf) (t / T * T)).
      {
        symmetry.
        rewrite vertical_map_canonical; omega.
      }
      rewrite <- H10.

      assert (H11: ZMap.get i (TAC_pool_at_t v (Z.to_nat N) conf) v <= ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t).
      { apply TAC_pool_at_t_monotonic_two_points; auto; try omega. }

      assert (H12: ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) (t / T * T) = ZMap.get i (TAC_pool_at_t (t/T*T) (Z.to_nat N) conf) (t / T * T)).
      {
        rewrite vertical_map_canonical; omega.
      }        
      rewrite H12 in Hfinish.
        
      destruct (zlt
                  (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t -
                   ZMap.get i (TAC_pool_at_t (t/T*T) (Z.to_nat N) conf) (t / T * T)) C); try discriminate.

      destruct (zlt
                  (ZMap.get i (TAC_pool_at_t v (Z.to_nat N) conf) v -
                   ZMap.get i (TAC_pool_at_t (t / T * T) (Z.to_nat N) conf) (t / T * T)) C); try omega.
      auto.
    }

    assert (Hisidle: highest_not_finished_map N v (TAC_pool_at_t v (Z.to_nat N) conf) conf
                                         (real_priqueue (Z.to_nat N) v conf) = N).
    {
      apply last_idle_is_idle with (ts := ts) (te:=te); auto.
    }

    assert (Hreal: 0 <= highest_not_finished_map N v (TAC_pool_at_t v (Z.to_nat N) conf) conf
                                         (real_priqueue (Z.to_nat N) v conf) < N).
    {    
      eapply exists_ready_task_highest_not_finished_map_schedule_something; eauto.
    }

    omega.
  Qed.
  
  Definition ty_not_possible_in_range ts te N conf id t1 ty :=
    forall t i T C,
      ts <= t < te ->
      i = highest_not_finished_map N t
                                   (TAC_pool_at_t t (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) t conf) ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      (get_sched_type i T id t t1) <> ty.    
  
  Lemma count_occurrence_ty_is_0_if_not_possible:
    forall N ts te conf id ty t1,
      0 <= ts <= te ->
      ty_not_possible_in_range ts te N conf id t1 ty ->
      count_type_occurrence N N ts te conf id t1 ty = 0.      
  Proof.
    intros * Ht Hty.
    unfold count_type_occurrence.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte in *.
    clear Heqnrange Hte.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: 0 <= ts <= ts + Z.of_nat nrange). omega.
      assert (H2: ty_not_possible_in_range ts (ts + Z.of_nat nrange) N conf id t1 ty).
      {
        unfold ty_not_possible_in_range in *.
        intros.
        apply Hty with (C:=C); auto.
        xomega.
      }
      rewrite (IHnrange H1 H2).

      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N); auto.
      destruct (ZMap.get
       (highest_not_finished_map N (ts + Z.of_nat nrange)
          (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) conf) as [T C] eqn:Hconfn.
      rewrite sched_type_eq_false; auto.

      unfold ty_not_possible_in_range in Hty.
      apply Hty with (C:=C); auto.
      xomega.
    }
  Qed.

  Lemma Sched_self_is_not_possible_since_v:
    forall N ts conf id T C v,
      0 <= id < N ->
      0 <= ts ->      
      1 < T ->
      ts mod T = 0 ->
      last_idle_slot N N ts (ts+T) conf = Some v ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      ty_not_possible_in_range (v + 1) (ts + T) N conf id (ts + T) Sched_self.
  Proof.
    intros * Hid Hts HT Htsmod Hidle Hconfid.
    assert (Hv: ts <= v < ts + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }

    unfold ty_not_possible_in_range.
    intros * Ht Hi Hconfi.
    destruct (zeq i id) as [Hself|Hother].
    {
      exfalso.
      rewrite <- Hself in *. clear Hself.

      assert (Hfinish: vt_map_not_finished t T C (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true).
      {
        apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:=conf); auto.
        - rewrite Z2Nat.id; auto. omega.
        - rewrite Z2Nat.id; auto. omega.
      }

      assert (H1: t / T * T + T = ts + T).
      {
        apply div_mul_round_up; try omega.
        rewrite mod_0_preserve; omega.
      }

      assert (Hv': v + 1 <= t / T * T).
      {
        apply (ready_task_must_be_within_last_idle N conf v ts (ts+T) t i T C); auto; try omega.
      }

      omega.
    }
    {
      unfold get_sched_type.
      rewrite zeq_false; auto.
      destruct (priEntry_lt (PriEntry (t / T0 * T0 + T0) i) (PriEntry (ts + T) id)); try congruence.
      destruct (zeq (t / T0 * T0 + T0) (ts + T)); try congruence.
    }
  Qed.
                           
  Lemma count_occurrence_self_is_0_since_v:
    forall N ts conf id T C v,
      0 <= id < N ->
      0 <= ts ->      
      1 < T ->
      ts mod T = 0 ->
      last_idle_slot N N ts (ts+T) conf = Some v ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      count_type_occurrence N N (v+1) (ts + T) conf id (ts+T) Sched_self = 0.
  Proof.
    intros.
    assert (Hv: ts <= v < ts + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }
    
    apply count_occurrence_ty_is_0_if_not_possible; auto; try omega. 
    - apply Sched_self_is_not_possible_since_v with (C:=C); auto.
  Qed.

End Sched_self.



Section Sched_interference.

  Definition single_task_ty_not_possible_in_range ts te N conf i T id t1 ty :=
    forall t,
      ts <= t < te ->
      i = highest_not_finished_map N t
                                   (TAC_pool_at_t t (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) t conf) ->
      (get_sched_type i T id t t1) <> ty.
  
  Lemma count_single_task_type_occurrence_ty_is_0_if_not_possible:
    forall N ts te conf i T C id ty t1,
      0 <= ts <= te ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      single_task_ty_not_possible_in_range ts te N conf i T id t1 ty ->
      count_single_task_type_occurrence N N i ts te conf id t1 ty = 0.      
  Proof.
    intros * Ht Hconfi Hty.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte in *.
    clear Heqnrange Hte.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: 0 <= ts <= ts + Z.of_nat nrange). omega.
      assert (H2: single_task_ty_not_possible_in_range ts (ts + Z.of_nat nrange) N conf i T id t1 ty).
      {
        unfold single_task_ty_not_possible_in_range in *.
        intros.
        apply Hty; auto.
        xomega.
      }
      rewrite (IHnrange H1 H2).

      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) i) as [Heq|Hneq]; auto.
      rewrite Heq in *.
      rewrite Hconfi.
      rewrite sched_type_eq_false; auto.

      unfold ty_not_possible_in_range in Hty.
      apply Hty; auto.
      xomega.
    }
  Qed.

  Lemma workload_remain_is_not_possible_since_v0:
    forall N ts conf id T C v i Ti Ci,
      0 <= ts ->
      1 < T ->
      1 < Ti ->
      0 <= i < N ->
      last_idle_slot N N ts (ts+T) conf = Some v ->
      (v+1) mod Ti <> 0 ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      single_task_ty_not_possible_in_range (v + 1) ((v+1)/Ti*Ti+Ti) N conf i Ti id (ts + T) Sched_interference.
  Proof.
    intros * Hts HT HTi Hi Hidle Hmod Hconfid Hconfi.
    assert (Hv: ts <= v < ts + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }

    unfold single_task_ty_not_possible_in_range.
    intros * Ht Hsched.

    unfold get_sched_type.
    destruct (zeq i id); try congruence.

    assert (Hfinish: vt_map_not_finished t Ti Ci (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true).
    {
      apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:=conf); auto.
      - rewrite Z2Nat.id; auto. omega.
      - rewrite Z2Nat.id; auto. omega.
    }
    
    destruct (priEntry_lt (PriEntry (t / Ti * Ti + Ti) i) (PriEntry (ts + T) id)) eqn:Hpri.
    {
      exfalso.
      
      assert (Hv': v + 1 <= t / Ti * Ti).
      {
        apply (ready_task_must_be_within_last_idle N conf v ts (ts+T) t i Ti Ci); auto; try omega.
      }

      assert (H1: 0 <= (v + 1) / Ti * Ti).
      {
        apply div_mul_nonneg; omega.
      }
      assert (H2: v + 1 - (v + 1) / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      assert (H3: ((v + 1) / Ti * Ti) mod Ti = 0).
      {
        apply Z_mod_mult.
      }
      
      assert (t / Ti * Ti = (v + 1) / Ti * Ti).
      {
        apply mod_lower_bound; try omega.
      }
      omega.
    }
    {
      destruct (zeq (t / Ti * Ti + Ti) (ts + T)); try congruence.
    }
  Qed.
  
  Lemma count_single_task_interference_bounded_by_workload_subtotal_v:
    forall N ts conf toid T C i Ti Ci v,
      0 < N ->
      0 <= ts ->
      last_idle_slot N N ts (ts+T) conf = Some v ->      
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      ZMap.get toid conf = mkPrioConfigValid T C ->
      1 < Ti <= T ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      count_single_task_type_occurrence N N i (v+1) (ts + T) conf toid (ts+T) Sched_interference <=
      workload_whole toid i Ti Ci (v+1) (ts + T).
  Proof.
    intros * HN Hts Hidle Hconf Hconftoid HTf HCf Hid Hibound.
    assert (Hv: ts <= v < ts + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }
    
    destruct (zeq ((v+1) mod Ti) 0) as [Hmod|Hmod].
    {
      apply count_single_task_interference_bounded_by_workload_subtotal_start_aligned_generalized; auto; try omega.
    }
    {
      assert (H1: v + 1 < (v+1) / Ti * Ti + Ti).
      {
        apply x_lt_floor_plus; omega.
      }
      assert (H2: v+1 - (v+1) / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }

      assert (Hty: single_task_ty_not_possible_in_range (v + 1) ((v+1)/Ti*Ti+Ti) N conf i Ti toid (ts + T) Sched_interference).
      {        
        apply workload_remain_is_not_possible_since_v0 with (C:=C) (Ci:=Ci); auto; omega.
      }
        
      destruct (zle ((v+1)/Ti*Ti+Ti) (ts + T)).
      {
        assert (H3: count_single_task_type_occurrence N N i (v+1) (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i (v+1) ((v+1) / Ti * Ti + Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i ((v+1) / Ti * Ti + Ti) (ts + T) conf toid (ts + T) Sched_interference).
        {
          symmetry.
          apply count_single_task_type_occurrence_composition; try omega.
        }
        rewrite H3. clear H3.
        
        assert (H4: count_single_task_type_occurrence N N i ((v+1) / Ti * Ti + Ti) 
                                                      (ts + T) conf toid (ts + T) Sched_interference <= workload_whole toid i Ti Ci ((v+1) / Ti * Ti + Ti) (ts + T)).
        {
          apply count_single_task_interference_bounded_by_workload_subtotal_start_aligned_generalized; auto; try omega.
          - rewrite Z.add_comm.
            rewrite Z_mod_plus_full.
            apply Z_mod_same_full.
        }

        rewrite workload_whole_left_align; try omega.

        assert (count_single_task_type_occurrence N N i (v + 1) 
                                                  ((v + 1) / Ti * Ti + Ti) conf toid (ts + T) Sched_interference = 0).
        {
          apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Ti) (C:=Ci); auto; try xomega.
        }
        omega.
      }
      {
        assert (H0: count_single_task_type_occurrence N N i (v + 1) (ts + T) conf toid (ts + T) Sched_interference = 0).
        {
          apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Ti) (C:=Ci); auto; try xomega.
          - unfold single_task_ty_not_possible_in_range in *.
            intros.
            apply Hty; auto.
            omega.
        }
        rewrite H0.
        apply workload_whole_nonneg; try omega.
      }
    }    
  Qed.

  Lemma count_occurrence_inteference_bounded_whole_aux:
    forall Nnat N0 id ts v T C conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts ->
      1 < T ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      last_idle_slot N0 N0 ts (ts+T) conf = Some v ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence2_aux Nnat N0 N0
                           (v+1) (ts + T) conf id (ts+T) Sched_interference <=
      workload_whole_total Nnat id (v+1) (ts+T) conf.
  Proof.
    intros * HNnat HN0 Hts HT Hconfid Hidle Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.

      assert (Hv: ts <= v < ts + T).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }      

      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.        
        rewrite count_single_task_self_sched_others_is_0; auto; try omega.
        congruence.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H3: Tf <= T).
        {
          apply Hlarge with (i:=Z.of_nat Nnat) (Cn:=Cf); auto.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }

        assert (H: forall a b c d, a <= d -> b <= c -> a + b <= c + d). intros. omega.
        apply H; auto.

        apply count_single_task_interference_bounded_by_workload_subtotal_v with (C:=C); auto.
        - omega.
      }
    }      
  Qed.

  Lemma count_occurrence_interference_bounded_by_whole:
    forall Nnat id ts T C conf v,
      0 < Z.of_nat Nnat ->
      0 <= ts ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) ts (ts+T) conf = Some v ->      
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                            (v+1) (ts + T) conf id (ts+T) Sched_interference <=
      workload_whole_total Nnat id (v+1) (ts + T) conf.
  Proof.
    intros * HN Hts Hid Hconfid Hconfvalid Hlarge.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    assert (HT: 1 < T).
    {
      specialize (Hconfvalid _ _ _ Hid Hconfid).
      omega.
    }
    assert (Hv: ts <= v < ts + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }          
    rewrite count_type_occurrence2_eq; try omega.
    rewrite H1.    
    
    apply count_occurrence_inteference_bounded_whole_aux with (C:=C); auto; try omega.     
  Qed.
  
End Sched_interference.



Section Sched_leftover.

  Lemma is_leftover_type_is_leftover_generalized_v:
    forall fromid Tf toid t ts te,
      0 <= ts ->
      1 < Tf ->
      te - Tf <= t < te ->
      is_leftover fromid Tf toid ts te = true ->
      get_sched_type fromid Tf toid t te = Sched_leftover.
  Proof.
    intros * Hts HTf Ht Hleftover.
    unfold is_leftover in Hleftover.
    destruct (zle ts (te - Tf)); try congruence.
    destruct (zeq (te mod Tf) 0) as [Hmod|Hmod]; try congruence.
    destruct (zlt toid fromid); try congruence.

    unfold get_sched_type.
    rewrite zeq_false; try omega.

    assert (H1: (te - Tf) mod Tf = 0).
    {
      replace te with ((te - Tf) + Tf) in Hmod by omega.
      rewrite Zplus_mod in Hmod.
      rewrite Z_mod_same_full in Hmod.
      replace ((te - Tf) mod Tf + 0) with ((te - Tf) mod Tf) in Hmod by omega.
      rewrite Zmod_mod in Hmod.
      auto.
    }
    assert (H: t / Tf * Tf + Tf = te).
    {
      remember (t - te + Tf) as d.
      replace t with ((te - Tf) + d) by omega.
      rewrite divide_plus_equal; try omega.
      - rewrite mod_0_divide_mul; try omega.
    }
    rewrite H.

    unfold priEntry_lt.
    rewrite zlt_false; try omega.
    repeat rewrite zeq_true.
    rewrite zlt_false; try omega.
    auto.
  Qed.

  Lemma is_leftover_yet_not_leftover_when_t_is_small_v:
    forall fromid Tf toid ts te t,
      0 <= ts ->      
      ts <= t < te - Tf ->
      1 < Tf ->
      is_leftover fromid Tf toid ts te = true ->
      get_sched_type fromid Tf toid t te <> Sched_leftover.
  Proof.
    intros * Hts Ht HTf Hleftover.
    unfold get_sched_type, not.
    intros H.    
    destruct (zeq fromid toid); try congruence.
    destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) fromid) (PriEntry te toid)); try congruence.
    destruct (zeq (t / Tf * Tf + Tf) te) as [Heq | Hneq]; try congruence.

    unfold is_leftover in Hleftover.
    destruct (zle ts (te - Tf)); try congruence.
    destruct (zeq (te mod Tf) 0) as [Hmod|Hmod]; try congruence.

    assert (t / Tf * Tf + Tf <= te - Tf).
    {
      apply mod_0_bound_increase; omega.
    }
    omega.
  Qed.
  
  Lemma is_leftover_then_single_task_type_count_is_0_with_small_t_v:
    forall N0 fromid toid ts te Tf Cf conf nrange,
      0 <= ts ->
      ts + Z.of_nat nrange <= te - Tf ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      is_leftover fromid Tf toid ts te = true ->
      count_single_task_type_occurrence N0 N0 fromid ts (ts+Z.of_nat nrange) conf toid te Sched_leftover = 0.
  Proof.
    intros * Hts Hrange Hconf HTf Hleftover.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.
      rewrite Heq.
      rewrite Hconf.
      rewrite sched_type_eq_false; auto.
      apply is_leftover_yet_not_leftover_when_t_is_small_v with (ts:=ts); auto.
      xomega.
    }
  Qed.
  
  Lemma count_single_task_self_sched_leftover_bounded_by_budget_aux_v:
    forall N ts te nrange conf i Ti Ci toid,
      0 <= N ->
      1 < Ti <= te ->
      0 < Ci ->
      0 <= ts ->      
      te - Ti < ts + Z.of_nat nrange <= te ->
      0 <= i < N ->      
      is_leftover i Ti toid ts te = true ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid te Sched_leftover = Z.min Ci
         (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (ts + Z.of_nat nrange) - (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti))).
  Proof.
    intros * HN HT HC Hts Hrange Hi Hleftover Hconf.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    
    induction nrange.
    {      
      simpl in *.
      replace (ts + 0) with ts by omega.
      unfold is_leftover in Hleftover.
      destruct (zle ts (te - Ti)); try congruence.
      omega.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H1.
      simpl.
      assert (HNN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.

      assert (H2: ts + Z.of_nat (S nrange) = ts + Z.of_nat nrange + 1).
      { rewrite Nat2Z.inj_succ; omega. }
      assert (H11: 0 <= i < Z.of_nat (Z.to_nat N)).
      { rewrite Z2Nat.id; omega. }
      assert (H20: te mod Ti = 0).
      {
        unfold is_leftover in Hleftover.
        destruct (zle ts (te - Ti)); try congruence.
        destruct (zeq (te mod Ti) 0); auto; try congruence.
      }
      assert (H21: (te - Ti) mod Ti = 0).
      {
        replace te with (te - Ti + Ti) in H20 by omega.
        rewrite Zplus_mod in H20.
        rewrite Z_mod_same_full in H20.
        replace ((te - Ti) mod Ti + 0) with ((te - Ti) mod Ti) in H20 by omega.
        rewrite Zmod_mod in H20.
        auto.
      }

      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) i) as [Hsched|Hother].
      {
        rewrite Hsched.
        rewrite Hconf.
        
        assert (Htype: get_sched_type i Ti toid (ts + Z.of_nat nrange) te = Sched_leftover).
        {
          apply is_leftover_type_is_leftover_generalized_v with (ts:=ts); auto; omega.
        }
        rewrite Htype.
        rewrite sched_type_eq_true.
        
        destruct (zeq (ts + Z.of_nat nrange) (te - Ti)) as [Heq | Hneq].
        {          
          rewrite Heq in *. clear Heq.

          assert (H0: count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid te Sched_leftover = 0).
          {
            apply is_leftover_then_single_task_type_count_is_0_with_small_t_v with (Tf:=Ti) (Cf:=Ci); auto; xomega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
          
          assert (H5: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti + 1) = ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti) + 1).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto; try omega.
            - rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          
          assert (H14: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) =
                       ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto; try omega.
            - rewrite Z2Nat.id; omega.            
          }

          rewrite H14.

          assert (H15: (ts + Z.of_nat nrange) / Ti * Ti = te - Ti).
          {
            apply mod_lower_bound; try omega.
          }

          assert (H16: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
      ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti) < Ci).
          {
            symmetry in Hsched.
            replace N with (Z.of_nat (Z.to_nat N)) in Hsched.
            
            rewrite Nat2Z.id in Hsched.
            pose proof (highest_not_finished_map_condition_met _ _ _ _ _ _ _ Hsched H11 Hconf).
            - unfold vt_map_not_finished in H.
              rewrite vertical_map_canonical; try omega.
              rewrite H15 in H.
              assert ((ts + Z.of_nat nrange) / Ti * Ti <= ts + Z.of_nat nrange).
              { apply divide_mul_le; omega. }
              
              assert (H17: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (te - Ti) = ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti)).
              {
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }
              rewrite H17 in H.
              destruct (zlt
        (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
           (ts + Z.of_nat nrange) -
         ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti)) Ci); try omega.
              discriminate.
          }
          repeat rewrite Z.min_r; try omega.
        }
      }
      {
        (* i not scheduled *)
        destruct (zeq (ts + Z.of_nat nrange) (te - Ti)) as [Heq | Hneq].
        {          
          rewrite Heq in *. clear Heq.
          
          assert (H0: count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid te Sched_leftover = 0).
          {
            apply is_leftover_then_single_task_type_count_is_0_with_small_t_v with (Tf:=Ti) (Cf:=Ci); auto; try xomega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.

          assert (H6:    vt_map_not_finished (te - Ti) Ti Ci
                                             (ZMap.get i (TAC_pool_at_t (te - Ti) (Z.to_nat N) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          assert (H5: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti + 1) = ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf) (te - Ti)).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Ti) (C:=Ci); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=Ti) (C:=Ci); auto.              
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.        
        }
        {
          rewrite IHnrange; try omega.
          destruct (vt_map_not_finished (ts + Z.of_nat nrange) Ti Ci
                                        (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
                                 (ts + Z.of_nat nrange + 1) =
                   ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
                            (ts + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Ti) (C:=Ci); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=Ti) (C:=Ci); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            auto.
          }
          {
            unfold vt_map_not_finished in Hfinish.            
            
            assert (H16: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
                                              (ts + Z.of_nat nrange) \/
                    ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
                                              (ts + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }
            
            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.

              assert (H18: ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)
        (te - Ti) = ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                                                      (te - Ti)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }

              rewrite H18.
              
              assert (H15: (ts + Z.of_nat nrange) / Ti * Ti = te - Ti).
              {
                apply mod_lower_bound; try omega.
              }
              rewrite H15 in Hfinish.
              
              destruct (zlt
              (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                 (ts + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                 (te - Ti)) Ci); try discriminate.

              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }
  Qed.    

  Lemma is_leftover_false_get_type_not_leftover_v:
    forall i Ti toid ts t te,
      0 < Ti ->
      is_leftover i Ti toid ts te = false ->
      ts <= te - Ti ->
      ts <= t <= te ->
      get_sched_type i Ti toid t te <> Sched_leftover.
  Proof.
    intros * HT Hleftover Hts Hrange.
    unfold not.
    intros H.
    unfold is_leftover in Hleftover.    
    unfold get_sched_type in H.

    destruct (zeq i toid); try congruence.
    destruct (priEntry_lt (PriEntry (t / Ti * Ti + Ti) i) (PriEntry te toid)) eqn:Hpri; try congruence.
    unfold priEntry_lt in Hpri.    
    destruct (zeq (t / Ti * Ti + Ti) te) as [Heq | Hneq]; try congruence.
    rewrite Heq in Hpri.
    rewrite zlt_false in Hpri; try omega.
    destruct (zlt i toid); try congruence.
    rewrite zle_true in Hleftover; auto.
    rewrite zlt_true in Hleftover; try omega.

    assert (Hmod: te mod Ti = 0).
    {
      rewrite <- Heq.
      rewrite mod_0_preserve; try omega.
      apply Z_mod_mult.
    }
    rewrite Hmod in Hleftover.
    rewrite zeq_true in Hleftover.
    congruence.
  Qed.

  Lemma leftover_not_possible_v:
    forall ts te N0 conf i Tf Cf id v,
      0 <= i < N0 ->
      0 <= ts <= te ->
      last_idle_slot N0 N0 ts te conf = Some v ->
      id <> i ->
      ZMap.get i conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      0 < Cf ->
      is_leftover i Tf id (v + 1) te = false ->
      single_task_ty_not_possible_in_range (v + 1) te N0 conf i Tf id te Sched_leftover.
  Proof.
    intros * Hi Hte Hidle Hid Hconfi HTf HCf Hleftover.
    unfold single_task_ty_not_possible_in_range.
    intros t Ht Hsched.

    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }      
    
    destruct (zle (v+1) (te - Tf)).
    {
      apply is_leftover_false_get_type_not_leftover_v with (ts := v+1); auto; try omega.
    }
    {
      assert (Hfinish: vt_map_not_finished t Tf Cf (ZMap.get i (TAC_pool_at_t t (Z.to_nat N0) conf)) = true).
      {
        apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N0) (conf:=conf); auto.
        - rewrite Z2Nat.id; auto. omega.
        - rewrite Z2Nat.id; auto. omega.
      }      
      
      assert (Hv': v + 1 <= t / Tf * Tf).
      {
        apply (ready_task_must_be_within_last_idle N0 conf v ts te t i Tf Cf); auto; try omega.
      }

      unfold get_sched_type.
      destruct (zeq i id); try congruence.
      destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) i) (PriEntry te id)); try congruence.
      destruct (zeq (t / Tf * Tf + Tf) te); try congruence.
      omega.
    }
  Qed.    
  
  Lemma count_occurrence_leftover_bounded_aux_v:
    forall Nnat N0 id ts te v conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts <= te ->
      last_idle_slot N0 N0 ts te conf = Some v ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->      
      count_type_occurrence2_aux Nnat N0 N0
                           (v+1) te conf id te Sched_leftover <=
      workload_leftover_total Nnat id (v+1) te conf.
  Proof.
    intros * HNnat HN0 Hts Hidle Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.

      assert (Hv: ts <= v < te).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }      
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_leftover_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        
        destruct (is_leftover (Z.of_nat Nnat) Tf id (v+1) te) eqn:Hleftover.
        {
          assert (H: forall a b c d, a <= d -> b <= c -> a + b <= c + d). intros. omega.
          apply H; auto.
          remember (Z.to_nat (te - (v+1))) as nT.
          assert (HH: te = v+1 + Z.of_nat nT).
          { rewrite HeqnT. rewrite Z2Nat.id; omega. }
          rewrite HH in *.
          rewrite count_single_task_self_sched_leftover_bounded_by_budget_aux_v with (Ti:=Tf) (Ci:=Cf); auto; try omega.
          - apply Z.le_min_l.
          - rewrite HeqnT. rewrite Z2Nat.id; auto; try omega.
            assert (H11: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
            specialize (Hlarge _ _ _ H11 Hconff).
            omega.
        }        
        {
          assert (H: count_single_task_type_occurrence N0 N0 (Z.of_nat Nnat) (v+1) 
                                                       te conf id te Sched_leftover = 0).
          {
            apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
            - omega.
            - apply leftover_not_possible_v with (ts:=ts) (Cf:=Cf); auto.
              omega.
          }
          rewrite H.
          omega.
        }
      }
    }      
  Qed.

  Lemma count_occurrence_leftover_bounded:
    forall Nnat id ts te v conf,
      0 < Z.of_nat Nnat ->
      0 <= ts <= te ->
      last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) ts te conf = Some v ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           (v+1) te conf id te Sched_leftover <=
      workload_leftover_total Nnat id (v+1) te conf.
  Proof.
    intros * HN Hts Hidle Hconfvalid Hlarge.
    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }      
    
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    apply count_occurrence_leftover_bounded_aux_v with (ts:=ts); auto; try omega. 
  Qed.
  
End Sched_leftover.


Section Sched_overlap.

  Lemma non_overlapping_no_remain_v:
    forall N0 conf fromid toid Tf Cf ts te te' v,
      last_idle_slot N0 N0 ts te conf = Some v ->
      0 <= fromid < N0 ->
      0 <= ts <= te ->
      te <= te' ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      0 < Cf ->
      is_overlapping fromid Tf toid (v+1) te te' = false ->
      workload_remain toid fromid Tf Cf te te' (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)) = 0.
  Proof.
    intros * Hidle Hfromid Hts Hte' Hconff HT HC Hnonoverlap.
    unfold is_overlapping in Hnonoverlap.
    unfold workload_remain.
    destruct (zeq (te mod Tf) 0); auto.
    assert (Hdestruct: (v + 1) > (te / Tf * Tf) \/ (te / Tf * Tf) >= te ->
           (if priEntry_lt (PriEntry (te / Tf * Tf + Tf) fromid) (PriEntry te' toid)
    then
     Z.max 0
       (Cf - ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) te +
        ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf))
            else 0) = 0).
    {
      intros o.
      assert (H1: te - te / Tf * Tf >= 1).
      {
        apply z_gt_mod; omega.
      }
      destruct o; try omega.

      assert (Hv: ts <= v < te).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }
      
      destruct (vt_map_not_finished te Tf Cf (ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N0) conf))) eqn:Hfinish.
      {
        exfalso.
        assert (Hv': v + 1 <= te / Tf * Tf).
        {
          eapply ready_task_must_be_within_last_idle; eauto; try omega.
        }
        omega.
      }
      {
        unfold vt_map_not_finished in Hfinish.
        destruct (priEntry_lt (PriEntry (te / Tf * Tf + Tf) fromid) (PriEntry te' toid)); auto.

        assert (Hvt1: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) te =
                      ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N0) conf) te).
        {
          rewrite vertical_map_canonical; omega.
        }
        rewrite Hvt1.

        assert (Hvt2: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf) =
                      ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N0) conf) (te / Tf * Tf)).
        {
          assert (te / Tf * Tf <= te).
          {
            apply divide_mul_le; omega.
          }
          assert (0 <= te / Tf * Tf).
          {
            apply div_mul_nonneg; omega.
          }
          rewrite vertical_map_canonical; try omega.
          symmetry.
          rewrite vertical_map_canonical; try omega.
        }
        rewrite Hvt2.

        destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N0) conf) te -
               ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N0) conf) (te / Tf * Tf)) Cf); try discriminate.
        rewrite Z.max_l; omega.
      }
    }
    destruct (zle (v + 1) (te / Tf * Tf)); try auto.
    destruct (zlt (te / Tf * Tf) te); try auto.
    {
      destruct (priEntry_lt (PriEntry (te / Tf * Tf + Tf) fromid) (PriEntry te' toid)); auto.
      congruence.
    }
  Qed.
  
  Lemma workload_overlapping_remain_total_eq_v:
    forall Nnat N0 id ts te te' v conf,
      0 <= ts <= te ->
      te <= te' ->
      Z.of_nat Nnat <= N0 ->
      last_idle_slot N0 N0 ts te conf = Some v ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
      ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      workload_overlapping_total Nnat id (v+1) te te' conf -
      workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf) =
      workload_overlapping_remain_total Nnat id (v+1) te te' te conf (TAC_pool_at_t te' (Z.to_nat N0) conf).
  Proof.
    intros * Hts Hte' HN Hidle Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (Hv: ts <= v < te).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }      
      
      assert (H3: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (H4: config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (H5: forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te).
      { intros. eapply Hlarge; eauto; omega. }
      assert (HN': Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat HN' H4 H5).
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.

      assert (H6: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
      assert (H7: 1 < Tn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      assert (H8: Tn <= te).
      { specialize (Hlarge _ _ _ H6 Hconfn). omega. }
      assert (H9: 0 < Cn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }

      destruct (is_overlapping (Z.of_nat Nnat) Tn id (v + 1) te te') eqn:Hoverlap.
      {
        rewrite is_overlapping_implies_remain with (d1:=v+1); auto; try omega.

        replace (Cn + workload_overlapping_total Nnat id (v + 1) te te' conf -
   (Cn - Z.min Cn (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) te - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tn * Tn)) +
    workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf))) with
            (workload_overlapping_total Nnat id (v + 1) te te' conf -
             workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf) +
            Z.min Cn
      (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) te - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tn * Tn))) by omega.

        rewrite <- IHNnat.
        auto.
      }
      {
        assert (Hremain: workload_remain id (Z.of_nat Nnat) Tn Cn te te' (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf)) = 0).
        {
          apply non_overlapping_no_remain_v with (ts:=ts) (v:=v); auto; try omega.          
        }
        omega.
      }
    }
  Qed.

  Lemma is_overlapping_type_is_overlap_generalized_v:
    forall fromid Tf toid t ts te te',
      1 < Tf  ->
      0 <= ts <= t ->
      te / Tf * Tf <= t < te ->
      is_overlapping fromid Tf toid ts te te' = true ->
      fromid <> toid ->
      get_sched_type fromid Tf toid t te = Sched_overlap.
  Proof.
    intros * HTf Hts Ht Hoverlap Hid.
    unfold is_overlapping in Hoverlap.
    unfold get_sched_type.
    rewrite zeq_false; auto.
    destruct (zle ts (te / Tf * Tf)); try congruence.
    destruct (zlt (te / Tf * Tf) te); try congruence.
    destruct (priEntry_lt (PriEntry (te / Tf * Tf + Tf) fromid)
                          (PriEntry te' toid)) eqn:Hpri; try congruence.

    unfold priEntry_lt.
    assert (H1: (te / Tf * Tf) mod Tf = 0). apply Z_mod_mult.
    assert (H2: te < te / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; omega.
    }
    assert (H3: te / Tf * Tf = (te / Tf * Tf ) / Tf * Tf).
    {
      symmetry.
      rewrite mod_0_divide_mul; try omega.    
    }
    
    assert (H4: te / Tf * Tf / Tf * Tf <= t / Tf * Tf).
    {
      apply div_mul_le; omega.
    }
    rewrite zlt_false; try omega.
    repeat rewrite zeq_false; try omega.
    auto.
  Qed.

  Lemma is_overlapping_not_possible_with_small_t:
    forall t te Tf N conf fromid toid,
      1 < Tf ->
      0 <= te ->
      0 <= t ->
      single_task_ty_not_possible_in_range t (te / Tf * Tf) N conf fromid Tf toid te Sched_overlap.
  Proof.
    intros * HT Hte Ht.
    unfold single_task_ty_not_possible_in_range.
    intros * Ht' Hfromid.
    unfold get_sched_type.
    destruct (zeq fromid toid); try congruence.
    destruct (priEntry_lt (PriEntry (t0 / Tf * Tf + Tf) fromid) (PriEntry te toid)) eqn:Hpri; try congruence.

    destruct (zeq (t0 / Tf * Tf + Tf) te); try congruence.
    unfold priEntry_lt in Hpri.
    destruct (zlt (t0 / Tf * Tf + Tf) te); try congruence.
    
    exfalso.
    assert (H1: te / Tf * Tf <= te).
    {
      apply divide_mul_le; try omega.
    }
    assert (H2: t0 / Tf * Tf + Tf <= te / Tf * Tf - 0 * Tf).
    {
      apply mod_0_bound_increase_k; try omega.
      apply Z_mod_mult.
    }
    omega.
  Qed.
  
  Lemma is_overlapping_then_single_task_type_count_is_valid_v:
    forall N0 fromid toid ts v te te' Tf Cf conf nrange,
      0 < N0 ->
      0 <= ts <= te ->
      te <= te' ->
      te / Tf * Tf < v+1 + Z.of_nat nrange <= te ->
      last_idle_slot N0 N0 ts te conf = Some v ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= te ->
      0 < Cf ->
      is_overlapping fromid Tf toid (v+1) te te' = true ->
      fromid <> toid ->
      0 <= fromid < N0 ->
      count_single_task_type_occurrence N0 N0 fromid (v+1) (v+1+ Z.of_nat nrange) conf toid te Sched_overlap = Z.min Cf (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (v+1 + Z.of_nat nrange) - ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) ((v+1 + Z.of_nat nrange) / Tf * Tf)).
  Proof.
    intros * HN0 Hts Hte' Hrange Hidle Hconf HTf HCf Hoverlap Hid Hfrom.
    unfold count_single_task_type_occurrence.
    replace (v+1 + Z.of_nat nrange - (v+1)) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl in Hrange.
      simpl in Hoverlap.
      unfold is_overlapping in Hoverlap.
      destruct (zle (v + 1) (te / Tf * Tf)); try omega; try congruence.
    }
    {
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      assert (H2: v+1 + Z.of_nat (S nrange) = (v+1 + Z.of_nat nrange) + 1). xomega.
      rewrite H2.
      assert (H3: (te / Tf * Tf) mod Tf = 0).
      { apply Z_mod_mult. }
      assert (H10: 0 <= te / Tf * Tf).
      {
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H11: 0 <= fromid < Z.of_nat (Z.to_nat N0)).
      { rewrite Z2Nat.id; omega. }
      assert (H12: (v+1 + Z.of_nat nrange + 1) mod Tf <> 0).
      {            
        apply a_between_floor_and_self_mod_not_0 with (c := te); omega.
      }
      assert (H15: 0 <= (v+1 + Z.of_nat nrange) / Tf * Tf).
      { 
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H20: (v+1 + Z.of_nat nrange) / Tf * Tf <= v+1 + Z.of_nat nrange).
      {
        apply divide_mul_le; omega.
      }
      assert (H21: te / Tf * Tf <= te).
      {
        apply divide_mul_le; omega.
      }
      
      simpl.

      assert (Hv: ts <= v < te).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }            

      destruct (zeq
         (highest_not_finished_map N0 (v + 1 + Z.of_nat nrange)
            (TAC_pool_at_t (v + 1 + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (v + 1 + Z.of_nat nrange) conf)) fromid) as [Hsched|Hnot].
      {
        rewrite Hsched.
        rewrite Hconf.
        
        assert (Htype: get_sched_type fromid Tf toid (v+1 + Z.of_nat nrange) te = Sched_overlap).
        {
          apply is_overlapping_type_is_overlap_generalized_v with (ts:=v+1) (te':=te'); auto; try omega.          
        }
        rewrite Htype.
        rewrite sched_type_eq_true.

        destruct (zeq (v+1 + Z.of_nat nrange) (te / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
            
          assert (H0: count_single_task_type_occurrence N0 N0 fromid (v+1) (v+1 + Z.of_nat nrange) conf toid te Sched_overlap = 0).
          {
            apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
            - omega.
            - rewrite Heq.
              apply is_overlapping_not_possible_with_small_t; auto; try omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (v+1 + Z.of_nat nrange - (v+1)) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
        
          assert (H4: (te / Tf * Tf + 1) / Tf * Tf = te / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto.
            rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.

          rewrite mod_equal; try omega.
          assert (H14: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (v+1 + Z.of_nat nrange + 1) =
                        ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (v+1 + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto.
            omega.
            rewrite Z2Nat.id; omega.            
          }

          rewrite H14.

          assert (H16: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (v+1 + Z.of_nat nrange) -
      ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
               ((v+1 + Z.of_nat nrange) / Tf * Tf) < Cf).
          {
            symmetry in Hsched.
            replace N0 with (Z.of_nat (Z.to_nat N0)) in Hsched.
            
            rewrite Nat2Z.id in Hsched.
            pose proof (highest_not_finished_map_condition_met _ _ _ _ _ _ _ Hsched H11 Hconf).
            - unfold vt_map_not_finished in H.
              rewrite vertical_map_canonical; try omega.
              assert ((v+1 + Z.of_nat nrange) / Tf * Tf <= v+1 + Z.of_nat nrange).
              { apply divide_mul_le; omega. }
              assert (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
     ((v+1 + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (v+1 + Z.of_nat nrange) (Z.to_nat N0) conf)
                                                   ((v+1 + Z.of_nat nrange) / Tf * Tf)).
              {
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }
              rewrite H4.
              destruct (zlt
        (ZMap.get fromid (TAC_pool_at_t (v + 1 + Z.of_nat nrange) (Z.to_nat N0) conf)
           (v + 1 + Z.of_nat nrange) -
         ZMap.get fromid (TAC_pool_at_t (v + 1 + Z.of_nat nrange) (Z.to_nat N0) conf)
           ((v + 1 + Z.of_nat nrange) / Tf * Tf)) Cf); try omega.
              discriminate H.
            - rewrite Z2Nat.id; omega.
          }

          repeat rewrite Z.min_r; omega.          
        }
      }
      {
        (* fromid not scheduled *)
        destruct (zeq (v+1 + Z.of_nat nrange) (te / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
          
          assert (H0: count_single_task_type_occurrence N0 N0 fromid (v+1) (v+1 + Z.of_nat nrange) conf toid te Sched_overlap = 0).
          {
            apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
            - omega.
            - rewrite Heq.
              apply is_overlapping_not_possible_with_small_t; auto; try omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (v+1 + Z.of_nat nrange - (v+1)) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
          
          assert (H4: (te / Tf * Tf + 1) / Tf * Tf = te / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H6: vt_map_not_finished (te / Tf * Tf) Tf Cf
                                          (ZMap.get fromid (TAC_pool_at_t (te / Tf * Tf) (Z.to_nat N0) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                               (te / Tf * Tf + 1) =
                      ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                               (te / Tf * Tf)).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          rewrite mod_equal; try omega.

          destruct (vt_map_not_finished (v+1 + Z.of_nat nrange) Tf Cf
     (ZMap.get fromid (TAC_pool_at_t (v+1 + Z.of_nat nrange) (Z.to_nat N0) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (v+1 + Z.of_nat nrange + 1) =
                   ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                            (v+1 + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            auto.
          }
          {
            unfold vt_map_not_finished in Hfinish.            
            
            assert (H16: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        (v+1 + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                              (v+1 + Z.of_nat nrange) \/
                    ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        (v+1 + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                              (v+1 + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }

            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.
              assert (H18: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        ((v+1 + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (v+1 + Z.of_nat nrange) (Z.to_nat N0) conf)
                                                      ((v+1 + Z.of_nat nrange) / Tf * Tf)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }

              rewrite H18.

              destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t (v + 1 + Z.of_nat nrange) (Z.to_nat N0) conf)
                 (v + 1 + Z.of_nat nrange) -
               ZMap.get fromid (TAC_pool_at_t (v + 1 + Z.of_nat nrange) (Z.to_nat N0) conf)
                 ((v + 1 + Z.of_nat nrange) / Tf * Tf)) Cf); try discriminate.

              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }
  Qed.      

  Lemma sched_overlap_not_possible2:
    forall i N Tf Cf v ts te te' conf id,
      last_idle_slot N N ts te conf = Some v ->
      is_overlapping i Tf id (v+1) te te' = false ->
      ZMap.get i conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= te ->
      0 <= ts <= te ->
      te + Tf <= te' ->
      0 <= i < N ->
      single_task_ty_not_possible_in_range (v+1) te N conf i Tf id te Sched_overlap.
  Proof.
    intros * Hidle Hoverlapping Hconfi Hte Hts Hlarge Hi.
    unfold single_task_ty_not_possible_in_range.
    intros * Ht Hsched.
    unfold is_overlapping in Hoverlapping.
    unfold get_sched_type.
    destruct (zeq i id); try congruence.
    destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) i) (PriEntry te id)) eqn:Hpri; try congruence.
    destruct (zeq (t / Tf * Tf + Tf) te); try congruence.
    exfalso.

    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }      

    assert (Hfinish: vt_map_not_finished t Tf Cf (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true).
    {
      apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:=conf); auto.
      - rewrite Z2Nat.id; auto. omega.
      - rewrite Z2Nat.id; auto; try omega.
    }

    assert (Hv': v + 1 <= t / Tf * Tf).
    {
      apply (ready_task_must_be_within_last_idle N conf v ts te t i Tf Cf); auto; try omega.
    }

    unfold priEntry_lt in Hpri.
    destruct (zlt (t / Tf * Tf + Tf) te); try congruence.

    destruct (zeq (te mod Tf) 0) as [Hmod|Hmod].
    {
      assert (H1: t / Tf * Tf + Tf <= te - 0 * Tf).
      {
        apply mod_0_bound_increase_k; omega.
      }
      nia.
    }
    {
      assert (H1: t / Tf * Tf <= te / Tf * Tf).
      {
        apply div_mul_le; omega.
      }
      assert (H2: te - te / Tf * Tf >= 1).
      {
        apply z_gt_mod; omega.
      }
      destruct (zle (v + 1) (te / Tf * Tf)); try omega.
      destruct (zlt (te / Tf * Tf) te); try omega.

      unfold priEntry_lt in Hoverlapping.
      rewrite zlt_true in Hoverlapping; try congruence.

      assert (H3: (te + Tf) / Tf * Tf <= te' / Tf * Tf).
      {
        apply div_mul_le; omega.
      }
      rewrite div_T_inc_1 in H3; try omega.
    }
  Qed.

  Lemma count_occurrence_overlapping_bounded_aux_v:
    forall Nnat N0 id ts te te' v conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 ->
      0 <= ts <= te ->
      te <= te' ->
      last_idle_slot N0 N0 ts te conf = Some v ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> te + Tn <= te') ->      
      count_type_occurrence2_aux Nnat N0 N0
                           (v+1) te conf id te Sched_overlap =
      workload_overlapping_remain_total Nnat id (v+1) te te' te conf (TAC_pool_at_t te' (Z.to_nat N0) conf).
  Proof.
    intros * HNnat HN0 Hts Hte' Hidle Hconfvalid Hlarge Hlarge2.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      assert (Hlarge2': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> te + Tn <= te').
      {
        intros.
        eapply Hlarge2; eauto.
        omega.
      }
      
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge' Hlarge2').
      simpl.
      rewrite <- IHNnat.

      assert (Hv: ts <= v < te).
      {
        eapply last_idle_slot_range; eauto.
        omega.
      }      
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_overlap_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.
        destruct (is_overlapping (Z.of_nat Nnat) Tf id (v+1) te te') eqn:Hoverlap.
        {
          assert (H: forall a b c, b = c -> a + b = a + c). intros. omega.
          apply H.
          remember (Z.to_nat (te - (v+1))) as nrange.
          assert (HTT: te = v+1+Z.of_nat nrange).
          { rewrite Heqnrange. rewrite Z2Nat.id; omega. }
          rewrite HTT in *. clear HTT Heqnrange.

          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.          
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: 0 < Cf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H5: Tf <= v+1+Z.of_nat nrange).
          { eapply Hlarge; eauto. }          

          apply is_overlapping_then_single_task_type_count_is_valid_v with (ts:=ts) (te:=v+1+Z.of_nat nrange); auto; try omega.
          - pose proof (is_overlapping_true_end_not_aligned _ _ _ _ _ _ Hoverlap).
            assert ((v+1 + Z.of_nat nrange) / Tf * Tf <= v+1 + Z.of_nat nrange).
            { apply divide_mul_le; omega. }
            omega.
        }
        {
          assert (H: forall a b, b = 0 -> a + b = a). intros. omega.
          apply H.
          remember (Z.to_nat (te - (v+1))) as nrange.
          assert (HTT: te = v+1+Z.of_nat nrange).
          { rewrite Heqnrange. rewrite Z2Nat.id; omega. }
          rewrite HTT in *. clear HTT Heqnrange.
          
          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: Tf <= v + 1 + Z.of_nat nrange).
          { eapply Hlarge; eauto. }

          apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
          - omega.
          - apply sched_overlap_not_possible2 with (ts:=ts) (te':=te') (Cf:=Cf); auto; try omega.
            assert (H30: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
            specialize (Hlarge2 _ _ _ H30 Hconff).
            omega.
        }
      }
    }
  Qed.

  Lemma count_occurrence_overlapping_bounded_v:
    forall Nnat id ts te te' v conf,
      0 < Z.of_nat Nnat ->
      0 <= ts <= te ->
      te <= te' ->
      last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) ts te conf = Some v ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> te + Tn <= te') ->      
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           (v+1) te conf id te Sched_overlap =
      workload_overlapping_remain_total Nnat id (v+1) te te' te conf (TAC_pool_at_t te' Nnat conf).
  Proof.
    intros * HN Hts Hte Hidle Hconfvalid Hlarge Hlarge2.

    assert (Hv: ts <= v < te).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }          
    
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    rewrite count_occurrence_overlapping_bounded_aux_v with (ts:=ts) (te':=te'); auto; try omega.
    - rewrite H1. auto.
  Qed.    

End Sched_overlap.


Section BUSY_SCHEDULE.

  Theorem interference_is_bounded_when_schedule_is_busy_last_idle:
    forall Nnat conf id T C np v,
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn : Z, 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) 
            (Z.of_nat np * T + T) conf = Some v ->
      workload_overlapping_total Nnat id (v+1) (Z.of_nat np * T + T) (Z.of_nat np * T + T + T) conf -
      workload_remain_total Nnat id (Z.of_nat (S np) * T) 
                            (Z.of_nat (S np) * T + T) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * T + T) Nnat conf) >=
      Z.of_nat np * T + T - (v+1) -
                        workload_leftover_total Nnat id (v+1) (Z.of_nat np * T + T) conf -
                        workload_whole_total Nnat id (v+1) (Z.of_nat np * T + T) conf.  
  Proof.
    intros * Hconfvalid Hlarge Hid Hconfid Hidle.
    assert (HTC: 1 < T /\ 0 < C).
    {
      specialize (Hconfvalid _ _ _ Hid Hconfid).
      omega.
    }
    destruct HTC as [HT HC].
    assert (Hts: 0 <= Z.of_nat np * T).
    { apply Zmult_gt_0_le_0_compat; xomega. }
    assert (Htsmod: (Z.of_nat np * T) mod T = 0).
    { rewrite Z_mod_mult; auto. }

    assert (Hidle2: last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) 
                                   (v+1) (Z.of_nat np * T + T) conf = None).
    {
      apply last_idle_slot_none with (ts:=Z.of_nat np * T); auto.
      - xomega.
      - omega.
    }
    assert (Hv: Z.of_nat np * T <= v < Z.of_nat np * T + T).
    {
      eapply last_idle_slot_range; eauto.
      omega.
    }
    
    assert (Hcount_total: count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (v+1) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_self + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (v+1) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_interference + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (v+1) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_leftover + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (v+1) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_overlap = Z.of_nat np * T + T - (v+1)).
    {
      apply sched_type_is_complete_when_schedule_is_busy_generalized with (T:=T) (C:=C); auto.
      omega.
    }

    assert (Hself: count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (v+1) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_self = 0).
    {
      apply count_occurrence_self_is_0_since_v with (C:=C); auto.
    }

    assert (Hinterference: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (v+1) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_interference <=
           workload_whole_total Nnat id (v+1) (Z.of_nat np * T + T) conf).
    {
      apply count_occurrence_interference_bounded_by_whole with (C:=C); auto.
      omega.
    }

    assert (Hleftover: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (v+1) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_leftover <=
                       workload_leftover_total Nnat id (v+1) (Z.of_nat np * T + T) conf).
    {
      apply count_occurrence_leftover_bounded with (ts:=Z.of_nat np * T); auto; try omega.
      - intros * Hi Hconfi.
        specialize (Hlarge _ _ _ Hi Hconfi).
        omega.
    }

    assert (Hoverlap: workload_overlapping_total Nnat id (v+1) (Z.of_nat np * T + T) (Z.of_nat np * T + T + T) conf -
   workload_remain_total Nnat id (Z.of_nat (S np) * T) 
                         (Z.of_nat (S np) * T + T) conf (TAC_pool_at_t (Z.of_nat (S np) * T + T) Nnat conf) =
           count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (v+1) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_overlap).
    {
      assert (H: Z.of_nat (S np) = Z.of_nat np + 1). xomega.
      rewrite H.
      replace ((Z.of_nat np + 1) * T) with (Z.of_nat np * T + T) by lia.

      assert (H2: workload_remain_total Nnat id (Z.of_nat np * T + T) 
                                        (Z.of_nat np * T + T + T) conf (TAC_pool_at_t (Z.of_nat np * T + T + T) Nnat conf) =
             workload_remain_total Nnat id (Z.of_nat np * T + T) 
                                   (Z.of_nat np * T + T + T) conf (TAC_pool_at_t (Z.of_nat np * T + T + T) (Z.to_nat (Z.of_nat Nnat)) conf)).
      {
        rewrite Nat2Z.id. auto.
      }
      rewrite H2.

      assert (H3: forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= Z.of_nat np * T + T).
      {
        intros * Hi Hconfi.
        specialize (Hlarge _ _ _ Hi Hconfi).
        omega.
      }
      rewrite workload_overlapping_remain_total_eq_v with (ts:=Z.of_nat np * T); auto; try xomega.
      symmetry.

      rewrite count_occurrence_overlapping_bounded_v with (ts:=Z.of_nat np * T) (te':=Z.of_nat np * T + T + T); auto; try omega.
      - rewrite Nat2Z.id. auto.
      - intros * Hi Hconfi.
        specialize (Hlarge _ _ _ Hi Hconfi).
        omega.
    }

    omega.
  Qed.    

End BUSY_SCHEDULE.


